-- Automatically generated
return { manifest = {
	components =
	{
		{
			name = "@sys.appcon.AVTA",
			location = "bin",
			depends    = {},
			provides   = {},
			parameters = { purge=false, autostart=true },
			version = "1.3.1.001",
		}
	}
}, build = "2020-04-22T21:23:08-0700", commit = "9392296" }
